import pyglet
from matplotlib import pyplot as plt
from pymunk.pyglet_util import DrawOptions
from skimage.transform import rescale, rotate
from skimage.util import random_noise
from joblib import Parallel, delayed
from skimage.morphology import opening
from tqdm import tqdm
import pandas as pd
from skimage import draw
from skimage.exposure import match_histograms, rescale_intensity
from SyMBac.PSF import get_fluorescence_kernel, get_phase_contrast_kernel
import os
import skimage
from skimage.segmentation import find_boundaries
from scipy.ndimage import gaussian_filter
import numpy as np
from skimage.color import rgb2gray
from numpy import fft
from PIL import Image
import copy

from SyMBac.cell import Cell
from SyMBac.general_drawing import convolve_rescale, make_images_same_shape, perc_diff, raster_cell, transform_func
from SyMBac.scene_functions import create_space, step_and_update
from SyMBac.trench_geometry import trench_creator


def get_trench_segments(space):
    """    
    A function which extracts the rigid body trench objects from the pymunk space object. Space object should be passed
    from the return value of the run_simulation() function
    
    Returns
    -------
    List of trench segment properties, later used to draw the trench.
    """
    trench_shapes = []
    for shape, body in zip(space.shapes, space.bodies):
        if body.body_type == 2:
            trench_shapes.append(shape)

    trench_segment_props = []
    for x in trench_shapes:
        trench_segment_props.append([x.bb, x.area, x.a, x.b])

    trench_segment_props = pd.DataFrame(trench_segment_props)
    trench_segment_props.columns = ["bb", "area", "a", "b"]
    main_segments = trench_segment_props.sort_values("area", ascending=False).iloc[0:2]
    return main_segments


def run_simulation(trench_length, trench_width, cell_max_length, cell_width, sim_length, pix_mic_conv, gravity,
                   phys_iters, max_length_var, width_var, save_dir):
    """
    Runs the rigid body simulation of bacterial growth based on a variety of parameters. Opens up a Pyglet window to
    display the animation in real-time. If the simulation looks bad to your eye, restart the kernel and rerun the
    simulation. There is currently a bug where if you try to rerun the simulation in the same kernel, it will be
    extremely slow.
    
    Parameters
    ----------

    trench_length : float
        Length of a mother machine trench (micron)
    trench_width : float
        Width of a mother machine trench (micron)
    cell_max_length : float
        Maximum length a cell can reach before dividing (micron)
    cell_width : float
        the average cell width in the simmulation (micron)
    pix_mic_conv : float
        The micron/pixel size of the image
    gravity : float
        Pressure forcing cells into the trench. Typically left at zero, but can be varied if cells start to fall into
        each other or if the simulation behaves strangely.
    phys_iters : int
        Number of physics iterations per simulation frame. Increase to resolve collisions if cells are falling into one
        another, but decrease if cells begin to repel one another too much (too high a value causes cells to bounce off
        each other very hard). 20 is a good starting point
    max_length_var : float
        Variance of the maximum cell length
    width_var : float
        Variance of the maximum cell width
    save_dir : str
        Location to save simulation outupt
        
    Returns
    -------
    cell_timeseries : lists
        A list of parameters for each cell, such as length, width, position, angle, etc. All used in the drawing of the
        scene later
    space : a pymunk space object
        Contains the rigid body physics objects which are the cells.
    """

    space = create_space()
    space.gravity = 0, gravity  # arbitrary units, negative is toward trench pole
    dt = 1 / 100  # time-step per frame
    pix_mic_conv = 1 / pix_mic_conv  # micron per pixel
    scale_factor = pix_mic_conv * 3  # resolution scaling factor

    trench_length = trench_length * scale_factor
    trench_width = trench_width * scale_factor
    trench_creator(trench_width, trench_length, (35, 0), space)  # Coordinates of bottom left corner of the trench

    cell1 = Cell(
        length=cell_max_length * scale_factor,
        width=cell_width * scale_factor,
        resolution=60,
        position=(20 + 35, 10),
        angle=0.8,
        space=space,
        dt=1 / 60,
        growth_rate_constant=1,
        max_length=cell_max_length * scale_factor,
        max_length_mean=cell_max_length * scale_factor,
        max_length_var=max_length_var * np.sqrt(scale_factor),
        width_var=width_var * np.sqrt(scale_factor),
        width_mean=cell_width * scale_factor,
        parent = None
    )


    window = pyglet.window.Window(700, 700, "SyMBac", resizable=True)
    options = DrawOptions()

    @window.event
    def on_draw():
        window.clear()
        space.debug_draw(options)

    global cell_timeseries
    global x

    try:
        del cell_timeseries
    except:
        pass
    try:
        del x
    except:
        pass

    x = [0]
    cell_timeseries = []
    cells = [cell1]
    pyglet.clock.schedule_interval(step_and_update, interval=dt, cells=cells, space=space, phys_iters=phys_iters,
                                   ylim=trench_length, cell_timeseries=cell_timeseries, x=x, sim_length=sim_length,
                                   save_dir=save_dir)
    pyglet.app.run()
    # window.close()
    # phys_iters = phys_iters
    # for x in tqdm(range(sim_length+250),desc="Simulation Progress"):
    #    cells = step_and_update(dt=dt, cells=cells, space=space, phys_iters=phys_iters,ylim=trench_length*1.1, cell_timeseries = cell_timeseries, x=x, sim_length = sim_length, save_dir = save_dir)
    #    if x > 250:
    #        cell_timeseries.append(deepcopy(cells))
    return cell_timeseries, space


def generate_PC_OPL(main_segments, offset, scene, mask, media_multiplier, cell_multiplier, device_multiplier,
                    y_border_expansion_coefficient, x_border_expansion_coefficient, fluorescence, defocus):
    """
    Takes a scene drawing, adds the trenches and colours all parts of the image to generate a first-order phase contrast
    image, uncorrupted (unconvolved) by the phase contrat optics. Also has a fluorescence parameter to quickly switch to
    fluorescence if you want.
    
    Parameters
    ----------
    main_segments : list
        A list of the trench segments, used for drawing the trench
    offset : int
        The same offset from the draw_scene function. Used to know the cell offset.
    scene : 2D numpy array
        A scene image
    mask : 2D numpy array
        The mask for the scene
    media_multiplier : float
        Intensity multiplier for media (the area between cells which isn't the device)
    cell_multiplier : float
        Intensity multiplier for cell
    device_multiplier : float
        Intensity multiplier for device
    y_border_expansion_coefficient : int
        Another offset-like argument. Multiplies the size of the image on each side by this value. 3 is a good starting
        value because you want the image to be relatively larger than the PSF which you are convolving over it.
    x_border_expansion_coefficient : int
        Another offset-like argument. Multiplies the size of the image on each side by this value. 3 is a good starting
        value because you want the image to be relatively larger than the PSF which you are convolving over it.
    fluorescence : bool
        If true converts image to a fluorescence (hides the trench and swaps to the fluorescence PSF).
    defocus : float
        Simulated optical defocus by convolving the kernel with a 2D gaussian of radius defocus.
        
    Returns
    -------
    expanded_scene : 2D numpy array
        A large (expanded on x and y axis) image of cells in a trench, but unconvolved. (The raw PC image before
        convolution)
    expanded_scene_no_cells : 2D numpy array
        Same as expanded_scene, except with the cells removed (this is necessary for later intensity tuning)
    expanded_mask : 2D numpy array
        The masks for the expanded scene
    """

    def get_OPL_image(main_segments, offset, scene, mask, media_multiplier, cell_multiplier, device_multiplier,
                      y_border_expansion_coefficient, x_border_expansion_coefficient, fluorescence, defocus):
        segment_1_top_left = (0 + offset, int(main_segments.iloc[0]["bb"][0] + offset))
        segment_1_bottom_right = (
            int(main_segments.iloc[0]["bb"][3] + offset), int(main_segments.iloc[0]["bb"][2] + offset))

        segment_2_top_left = (0 + offset, int(main_segments.iloc[1]["bb"][0] + offset))
        segment_2_bottom_right = (
            int(main_segments.iloc[1]["bb"][3] + offset), int(main_segments.iloc[1]["bb"][2] + offset))

        if fluorescence:
            test_scene = np.zeros(scene.shape)
            media_multiplier = -1 * device_multiplier
        else:
            test_scene = np.zeros(scene.shape) + device_multiplier
            rr, cc = draw.rectangle(start=segment_1_top_left, end=segment_1_bottom_right, shape=test_scene.shape)
            test_scene[rr, cc] = 1 * media_multiplier
            rr, cc = draw.rectangle(start=segment_2_top_left, end=segment_2_bottom_right, shape=test_scene.shape)
            test_scene[rr, cc] = 1 * media_multiplier
            circ_midpoint_y = (segment_1_top_left[1] + segment_2_bottom_right[1]) / 2
            radius = (segment_1_top_left[1] - offset - (segment_2_bottom_right[1] - offset)) / 2
            circ_midpoint_x = (offset) + radius

            rr, cc = draw.rectangle(start=segment_2_top_left, end=(circ_midpoint_x, segment_1_top_left[1]),
                                    shape=test_scene.shape)
            test_scene[rr.astype(int), cc.astype(int)] = 1 * media_multiplier
            rr, cc = draw.disk(center=(circ_midpoint_x, circ_midpoint_y), radius=radius, shape=test_scene.shape)
            rr_semi = rr[rr < (circ_midpoint_x + 1)]
            cc_semi = cc[rr < (circ_midpoint_x + 1)]
            test_scene[rr_semi, cc_semi] = device_multiplier
        no_cells = copy.deepcopy(test_scene)

        test_scene += scene * cell_multiplier
        if fluorescence:
            pass
        else:
            test_scene = np.where(no_cells != media_multiplier, test_scene, media_multiplier)
        test_scene = test_scene[segment_2_top_left[0]:segment_1_bottom_right[0],
                     segment_2_top_left[1]:segment_1_bottom_right[1]]

        mask = np.where(no_cells != media_multiplier, mask, 0)
        mask_resized = mask[segment_2_top_left[0]:segment_1_bottom_right[0],
                       segment_2_top_left[1]:segment_1_bottom_right[1]]

        no_cells = no_cells[segment_2_top_left[0]:segment_1_bottom_right[0],
                   segment_2_top_left[1]:segment_1_bottom_right[1]]
        expanded_scene_no_cells = np.zeros((int(no_cells.shape[0] * y_border_expansion_coefficient),
                                            int(no_cells.shape[1] * x_border_expansion_coefficient))) + media_multiplier
        expanded_scene_no_cells[expanded_scene_no_cells.shape[0] - no_cells.shape[0]:,
        int(expanded_scene_no_cells.shape[1] / 2 - int(test_scene.shape[1] / 2)):int(
            expanded_scene_no_cells.shape[1] / 2 - int(test_scene.shape[1] / 2)) + no_cells.shape[1]] = no_cells
        if fluorescence:
            expanded_scene = np.zeros((int(test_scene.shape[0] * y_border_expansion_coefficient),
                                       int(test_scene.shape[1] * x_border_expansion_coefficient)))
            expanded_scene[expanded_scene.shape[0] - test_scene.shape[0]:,
            int(expanded_scene.shape[1] / 2 - int(test_scene.shape[1] / 2)):int(
                expanded_scene.shape[1] / 2 - int(test_scene.shape[1] / 2)) + test_scene.shape[1]] = test_scene
        else:
            expanded_scene = np.zeros((int(test_scene.shape[0] * y_border_expansion_coefficient),
                                       int(test_scene.shape[1] * x_border_expansion_coefficient))) + media_multiplier
            expanded_scene[expanded_scene.shape[0] - test_scene.shape[0]:,
            int(expanded_scene.shape[1] / 2 - int(test_scene.shape[1] / 2)):int(
                expanded_scene.shape[1] / 2 - int(test_scene.shape[1] / 2)) + test_scene.shape[1]] = test_scene

        expanded_mask = np.zeros((int(test_scene.shape[0] * y_border_expansion_coefficient),
                                  int(test_scene.shape[1] * x_border_expansion_coefficient)))
        expanded_mask[expanded_mask.shape[0] - test_scene.shape[0]:,
        int(expanded_mask.shape[1] / 2 - int(test_scene.shape[1] / 2)):int(
            expanded_mask.shape[1] / 2 - int(test_scene.shape[1] / 2)) + test_scene.shape[1]] = mask_resized

        return expanded_scene, expanded_scene_no_cells, expanded_mask

    expanded_scene, expanded_scene_no_cells, expanded_mask = get_OPL_image(main_segments, offset, scene, mask,
                                                                           media_multiplier, cell_multiplier,
                                                                           device_multiplier,
                                                                           y_border_expansion_coefficient,
                                                                           x_border_expansion_coefficient, fluorescence,
                                                                           defocus)
    if expanded_scene is None:
        main_segments = main_segments.reindex(index=main_segments.index[::-1])
        expanded_scene, expanded_scene_no_cells, expanded_mask = get_OPL_image(main_segments, offset, scene, mask,
                                                                               media_multiplier, cell_multiplier,
                                                                               device_multiplier,
                                                                               y_border_expansion_coefficient,
                                                                               x_border_expansion_coefficient,
                                                                               fluorescence, defocus)
    return expanded_scene, expanded_scene_no_cells, expanded_mask


def generate_test_comparison(media_multiplier=75, cell_multiplier=1.7, device_multiplier=29, sigma=8.85, scene_no=0,
                             scale=None, match_fourier=False, match_histogram=True, match_noise=False, offset=30,
                             debug_plot=False, noise_var=0.001, main_segments=None, scenes=None, kernel_params=None,
                             resize_amount=None, real_image=None, image_params=None, error_params=None,
                             x_border_expansion_coefficient=None, y_border_expansion_coefficient=None,
                             fluorescence=False, defocus=3.0):
    """

    Takes all the parameters we've defined and calculated, and uses them to finally generate a synthetic image. 
    
    Parameters
    ----------
    media_multiplier : float
        Intensity multiplier for media (the area between cells which isn't the device)
    cell_multiplier : float
        Intensity multiplier for cell
    device_multiplier : float
        Intensity multiplier for device
    sigma : float
        Radius of a gaussian which simulates PSF apodisation
    scene_no : int in range(len(cell_timeseries_properties))
        The index of which scene to render
    scale : float
        The micron/pixel value of the image
    match_fourier : bool
        If true, use sfmatch to match the rotational fourier spectrum of the synthetic image to a real image sample
    match_histogram : bool
        If true, match the intensity histogram of a synthetic image to a real image
    offset : int
        The same offset value from draw_scene
    debug_plot : bool
        True if you want to see a quick preview of the rendered synthetic image
    noise_var : float
        The variance for the simulated camera noise (gaussian)
    main_segments : list
        List of trench segment properties, output of get_trench_segments function
    scenes : list(2D numpy array)
        A list of the previously rendered scene mask pairs
    kernel_params : tuple
        A tuple of kernel parameters in this order: (R,W,radius,scale,NA,n,sigma,λ)
    resize_amount : int
        The upscaling factor to render the image by. E.g a resize_amount of 3 will interally render the image at 3x
        resolution before convolving and then downsampling the image. Values >2 are recommended.
    real_image : 2D numpy array
        A sample real image from the experiment you are trying to replicate
    image_params : tuple
        A tuple of parameters which describe the intensities and variances of the real image, in this order:
        (real_media_mean, real_cell_mean, real_device_mean, real_means, real_media_var, real_cell_var, real_device_var,
            real_vars).
    error_params : tuple
        A tuple of parameters which characterises the error between the intensities in the real image and the synthetic
        image, in this order: (mean_error,media_error,cell_error,device_error,mean_var_error,media_var_error,
        cell_var_error,device_var_error). I have given an example of their calculation in the example notebooks.
    y_border_expansioon_coefficient : int
        Another offset-like argument. Multiplies the size of the image on each side by this value. 3 is a good starting
        value because you want the image to be relatively larger than the PSF which you are convolving over it.
    x_border_expansioon_coefficient : int
        Another offset-like argument. Multiplies the size of the image on each side by this value. 3 is a good starting
        value because you want the image to be relatively larger than the PSF which you are convolving over it.
    fluorescence : bool
        If true converts image to a fluorescence (hides the trench and swaps to the fluorescence PSF). 
    defocus : float
        Simulated optical defocus by convolving the kernel with a 2D gaussian of radius defocus.
        
    Returns
    -------
    noisy_img : 2D numpy array
        The final simulated microscope image
    expanded_mask_resized_reshaped : 2D numpy array
        The final image's accompanying masks

    """

    expanded_scene, expanded_scene_no_cells, expanded_mask = generate_PC_OPL(
        main_segments=main_segments,
        offset=offset,
        scene=scenes[scene_no][0],
        mask=scenes[scene_no][1],
        media_multiplier=media_multiplier,
        cell_multiplier=cell_multiplier,
        device_multiplier=device_multiplier,
        x_border_expansion_coefficient=x_border_expansion_coefficient,
        y_border_expansion_coefficient=y_border_expansion_coefficient,
        fluorescence=fluorescence,
        defocus=defocus
    )

    R, W, radius, scale, NA, n, _, λ = kernel_params

    real_media_mean, real_cell_mean, real_device_mean, real_means, real_media_var, real_cell_var, real_device_var, real_vars = image_params
    mean_error, media_error, cell_error, device_error, mean_var_error, media_var_error, cell_var_error, device_var_error = error_params

    if fluorescence:
        kernel = get_fluorescence_kernel(radius=radius, scale=scale, NA=NA, n=n, Lambda=λ)[0]
        if defocus > 0:
            kernel = gaussian_filter(kernel, defocus, mode="reflect")
    else:
        kernel = get_phase_contrast_kernel(R=R, W=W, radius=radius, scale=scale, NA=NA, n=n, sigma=sigma, λ=λ)
        if defocus > 0:
            kernel = gaussian_filter(kernel, defocus, mode="reflect")

    convolved = convolve_rescale(expanded_scene, kernel, 1 / resize_amount, rescale_int=True)
    real_resize, expanded_resized = make_images_same_shape(real_image, convolved, rescale_int=True)
    fftim1 = fft.fftshift(fft.fft2(real_resize))
    angs, mags = cart2pol(np.real(fftim1), np.imag(fftim1))

    if match_fourier and not match_histogram:
        matched = sfMatch([real_resize, expanded_resized], tarmag=mags)[1]
        matched = lumMatch([real_resize, matched], None, [np.mean(real_resize), np.std(real_resize)])[1]
    else:
        matched = expanded_resized

    if match_histogram and match_fourier:
        matched = sfMatch([real_resize, matched], tarmag=mags)[1]
        matched = lumMatch([real_resize, matched], None, [np.mean(real_resize), np.std(real_resize)])[1]
        matched = match_histograms(matched, real_resize, multichannel=False)
    else:
        pass
    if match_histogram:
        matched = match_histograms(matched, real_resize, multichannel=False)
    else:
        pass

    noisy_img = random_noise(rescale_intensity(matched), mode="poisson")
    noisy_img = random_noise(rescale_intensity(noisy_img), mode="gaussian", mean=0, var=noise_var, clip=False)

    if match_noise:
        noisy_img = match_histograms(noisy_img, real_resize, multichannel=False)
    else:
        pass
    noisy_img = rescale_intensity(noisy_img.astype(np.float32), out_range=(0, 1))

    ## getting the cell mask to the right shape
    expanded_mask_resized = rescale(expanded_mask, 1 / resize_amount, anti_aliasing=False, preserve_range=True, order=0)
    if len(np.unique(expanded_mask_resized)) > 2:
        _, expanded_mask_resized_reshaped = make_images_same_shape(real_image, expanded_mask_resized, rescale_int=False)
    else:
        _, expanded_mask_resized_reshaped = make_images_same_shape(real_image, expanded_mask_resized, rescale_int=True)

    expanded_media_mask = rescale(
        (expanded_scene_no_cells == device_multiplier) ^ (expanded_scene - expanded_scene_no_cells).astype(bool),
        1 / resize_amount, anti_aliasing=False)
    real_resize, expanded_media_mask = make_images_same_shape(real_image, expanded_media_mask, rescale_int=True)
    just_media = expanded_media_mask * noisy_img

    expanded_cell_pseudo_mask = (expanded_scene - expanded_scene_no_cells).astype(bool)
    expanded_cell_pseudo_mask = rescale(expanded_cell_pseudo_mask, 1 / resize_amount, anti_aliasing=False)

    real_resize, expanded_cell_pseudo_mask = make_images_same_shape(real_image, expanded_cell_pseudo_mask,
                                                                    rescale_int=True)
    just_cells = expanded_cell_pseudo_mask * noisy_img

    expanded_device_mask = expanded_scene_no_cells

    expanded_device_mask = rescale(expanded_device_mask, 1 / resize_amount, anti_aliasing=False)
    real_resize, expanded_device_mask = make_images_same_shape(real_image, expanded_device_mask, rescale_int=True)
    just_device = expanded_device_mask * noisy_img

    simulated_means = np.array([just_media[np.where(just_media)].mean(), just_cells[np.where(just_cells)].mean(),
                                just_device[np.where(just_device)].mean()])
    simulated_vars = np.array([just_media[np.where(just_media)].var(), just_cells[np.where(just_cells)].var(),
                               just_device[np.where(just_device)].var()])
    mean_error.append(perc_diff(np.mean(noisy_img), np.mean(real_resize)))
    mean_var_error.append(perc_diff(np.var(noisy_img), np.var(real_resize)))
    if fluorescence:
        pass
    else:
        media_error.append(perc_diff(simulated_means[0], real_media_mean))
        cell_error.append(perc_diff(simulated_means[1], real_cell_mean))
        device_error.append(perc_diff(simulated_means[2], real_device_mean))

        media_var_error.append(perc_diff(simulated_vars[0], real_media_var))
        cell_var_error.append(perc_diff(simulated_vars[1], real_cell_var))
        device_var_error.append(perc_diff(simulated_vars[2], real_device_var))
    if debug_plot:
        fig = plt.figure(figsize=(15, 5))
        ax1 = plt.subplot2grid((1, 8), (0, 0), colspan=1, rowspan=1)
        ax2 = plt.subplot2grid((1, 8), (0, 1), colspan=1, rowspan=1)
        ax3 = plt.subplot2grid((1, 8), (0, 2), colspan=3, rowspan=1)
        ax4 = plt.subplot2grid((1, 8), (0, 5), colspan=3, rowspan=1)
        ax1.imshow(noisy_img, cmap="Greys_r")
        ax1.set_title("Synthetic")
        ax1.axis("off")
        ax2.imshow(real_resize, cmap="Greys_r")
        ax2.set_title("Real")
        ax2.axis("off")
        ax3.plot(mean_error)
        ax3.plot(media_error)
        ax3.plot(cell_error)
        ax3.plot(device_error)
        ax3.legend(["Mean error", "Media error", "Cell error", "Device error"])
        ax3.set_title("Intensity Error")

        ax4.plot(mean_var_error)
        ax4.plot(media_var_error)
        ax4.plot(cell_var_error)
        ax4.plot(device_var_error)
        ax4.legend(["Mean error", "Media error", "Cell error", "Device error"])
        ax4.set_title("Variance Error")

        fig.tight_layout()
        plt.show()
        plt.close()
    else:
        return noisy_img, expanded_mask_resized_reshaped.astype(int)


def draw_scene(cell_properties, do_transformation, space_size, offset, label_masks):
    """
    Draws a raw scene (no trench) of cells, and returns accompanying masks for training data.
    
    Parameters
    ----------
    cell properties : list
        A list of cell properties for that frame
    do_transformation : bool
        True if you want cells to be bent, false and cells remain straight as in the simulation
    space_size : tuple
        The xy size of the numpy array in which the space is rendered. If too small then cells will not fit. recommend using the get_space_size() function to find the correct space size for your simulation
    offset : int
        A necessary parameter which offsets the drawing a number of pixels from the left hand side of the image. 30 is a good number, but if the cells are very thick, then might need increasing. 
    label_masks : bool
        If true returns cell masks which are labelled (good for instance segmentation). If false returns binary masks only. I recommend leaving this as True, because you can always binarise the masks later if you want. 
        
    Returns
    -------
    space, space_masks : 2D numpy array, 2D numpy array
    
    space : 2D numpy array
        Not to be confused with the pyglet object calledspace in some other functions. Simply a 2D numpy array with an image of cells from the input frame properties
    space_masks : 2D numy array
        The masks (labelled or bool) for that scene.
    
    """
    space_size = np.array(space_size)  # 1000, 200 a good value
    space = np.zeros(space_size)
    space_masks_label = np.zeros(space_size)
    space_masks_nolabel = np.zeros(space_size)
    colour_label = [1]

    for properties in cell_properties:
        length, width, angle, position, freq_modif, amp_modif, phase_modif, phase_mult = properties
        position = np.array(position)
        x = np.array(position).astype(int)[0] + offset
        y = np.array(position).astype(int)[1] + offset
        OPL_cell = raster_cell(length=length, width=width)

        if do_transformation:
            OPL_cell_2 = np.zeros((OPL_cell.shape[0], int(OPL_cell.shape[1] * 2)))
            midpoint = int(np.median(range(OPL_cell_2.shape[1])))
            OPL_cell_2[:,
            midpoint - int(OPL_cell.shape[1] / 2):midpoint - int(OPL_cell.shape[1] / 2) + OPL_cell.shape[1]] = OPL_cell
            roll_coords = np.array(range(OPL_cell_2.shape[0]))
            freq_mult = (OPL_cell_2.shape[0])
            amp_mult = OPL_cell_2.shape[1] / 10
            sin_transform_cell = transform_func(amp_modif, freq_modif, phase_modif)
            roll_amounts = sin_transform_cell(roll_coords, amp_mult, freq_mult, phase_mult)
            for B in roll_coords:
                OPL_cell_2[B, :] = np.roll(OPL_cell_2[B, :], roll_amounts[B])
            OPL_cell = (OPL_cell_2)

        rotated_OPL_cell = rotate(OPL_cell, -angle, resize=True, clip=False, preserve_range=True, center=(x, y))
        cell_y, cell_x = (np.array(rotated_OPL_cell.shape) / 2).astype(int)
        offset_y = rotated_OPL_cell.shape[0] - space[y - cell_y:y + cell_y, x - cell_x:x + cell_x].shape[0]
        offset_x = rotated_OPL_cell.shape[1] - space[y - cell_y:y + cell_y, x - cell_x:x + cell_x].shape[1]
        assert y > cell_y, "Cell has {} negative pixels in y coordinate, try increasing your offset".format(y - cell_y)
        assert x > cell_x, "Cell has negative pixels in x coordinate, try increasing your offset"
        space[
        y - cell_y:y + cell_y + offset_y,
        x - cell_x:x + cell_x + offset_x
        ] += rotated_OPL_cell

        def get_mask(label_masks):

            if label_masks:
                space_masks_label[y - cell_y:y + cell_y + offset_y, x - cell_x:x + cell_x + offset_x] += (
                                                                                                                 rotated_OPL_cell > 0) * \
                                                                                                         colour_label[0]
                colour_label[0] += 1
                return space_masks_label
            else:
                space_masks_nolabel[y - cell_y:y + cell_y + offset_y, x - cell_x:x + cell_x + offset_x] += (
                                                                                                                   rotated_OPL_cell > 0) * 1
                return space_masks_nolabel
                # space_masks = opening(space_masks,np.ones((2,11)))

        label_mask = get_mask(True).astype(int)
        nolabel_mask = get_mask(False).astype(int)
        label_mask_fixed = np.where(nolabel_mask > 1, 0, label_mask)
        mask_borders = find_boundaries(label_mask_fixed, mode="thick", connectivity=2)
        space_masks = np.where(mask_borders, 0, label_mask_fixed)
        space_masks = opening(space_masks)
        if label_masks == False:
            space_masks = space_masks.astype(bool)
        space = space * space_masks.astype(bool)
    return space, space_masks


def generate_training_data(interactive_output, sample_amount, randomise_hist_match, randomise_noise_match, sim_length,
                           burn_in, n_samples, save_dir):
    """
    Generates the training data from a Jupyter interactive output of generate_test_comparison
    
    Parameters
    ----------
    interactive_output : ipywidgets.widgets.interaction.interactive
        The slider object after you have finished tweaking parameters
    sample_amount : float
        The percentage sampling variance (drawn from a uniform distribution) to vary intensities by. For example, a
        sample_amount of 0.05 will randomly sample +/- 5% above and below the chosen intensity for cells,
        media and device. Can be used to create a little bit of variance in the final training data.
    randomise_hist_match : bool
        If true, histogram matching is randomly turned on and off each time a training sample is generated
    randomise_noise_match : bool
        If true, noise matching is randomly turned on and off each time a training sample is generated
    sim_length : int
        the length of the simulation which was run
    burn_in : int
        Number of frames to wait before generating training data. Can be used to ignore the start of the simulation
        where the trench only has 1 cell in it.
    n_samples : int
        The number of training images to generate
    save_dir : str
        The save directory of the training data
    
    """

    media_multiplier, cell_multiplier, device_multiplier, sigma, scene_no, scale, match_fourier, match_histogram, \
    match_noise, offset, debug_plot, noise_var, main_segments, scenes, kernel_params, resize_amount, real_image, \
    image_params, error_params, x_border_expansion_coefficient, y_border_expansion_coefficient, fluorescence, \
    defocus = list(
        interactive_output.kwargs.values())
    debug_plot = False
    try:
        os.mkdir(save_dir)
    except:
        pass
    try:
        os.mkdir(save_dir + "/convolutions")
    except:
        pass
    try:
        os.mkdir(save_dir + "/masks")
    except:
        pass

    current_file_num = len(os.listdir(save_dir + "/convolutions"))

    def generate_samples(z):
        _media_multiplier = np.random.uniform(1 - sample_amount, 1 + sample_amount) * media_multiplier
        _cell_multiplier = np.random.uniform(1 - sample_amount, 1 + sample_amount) * cell_multiplier
        _device_multiplier = np.random.uniform(1 - sample_amount, 1 + sample_amount) * device_multiplier
        _sigma = np.random.uniform(1 - sample_amount, 1 + sample_amount) * sigma
        _scene_no = np.random.randint(burn_in, sim_length - 2)
        if randomise_hist_match:
            _match_histogram = np.random.choice([True, False])
        else:
            _match_histogram = match_histogram
        if randomise_noise_match:
            _match_noise = np.random.choice([True, False])
        else:
            _match_noise = match_noise

        syn_image, mask = generate_test_comparison(_media_multiplier, _cell_multiplier, _device_multiplier, _sigma,
                                                   _scene_no, scale, match_fourier, _match_histogram, _match_noise,
                                                   offset, debug_plot, noise_var, main_segments, scenes, kernel_params,
                                                   resize_amount, real_image, image_params, error_params,
                                                   x_border_expansion_coefficient, y_border_expansion_coefficient,
                                                   fluorescence, defocus)

        syn_image = Image.fromarray(skimage.img_as_uint(rescale_intensity(syn_image)))
        syn_image.save("{}/convolutions/synth_{}.tif".format(save_dir, str(z).zfill(5)))

        if (cell_multiplier == 0) or (cell_multiplier == 0.0):
            mask = np.zeros(mask.shape)
            mask = Image.fromarray(mask.astype(np.uint8))
            mask.save("{}/masks/synth_{}.tif".format(save_dir, str(z).zfill(5)))
        else:
            mask = Image.fromarray(mask.astype(np.uint8))
            mask.save("{}/masks/synth_{}.tif".format(save_dir, str(z).zfill(5)))
            ## TODO: change parallel if not using GPU

    Parallel(n_jobs=1)(delayed(generate_samples)(z) for z in
                       tqdm(range(current_file_num, n_samples + current_file_num), desc="Sample generation"))


# from https://stackoverflow.com/questions/20924085/python-conversion-between-coordinates
def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return phi, rho


def pol2cart(phi, rho):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return x, y


def sfMatch(images, rescaling=0, tarmag=None):
    assert type(images) == type([]), 'The input must be a list.'

    numin = len(images)
    xs, ys = images[1].shape
    angs = np.zeros((xs, ys, numin))
    mags = np.zeros((xs, ys, numin))
    for x in range(numin):
        if len(images[x].shape) == 3:
            images[x] = rgb2gray(images[x])
        im1 = images[x] / 255
        xs1, ys1 = im1.shape
        assert (xs == xs1) and (ys == ys1), 'All images must have the same size.'
        fftim1 = fft.fftshift(fft.fft2(im1))
        angs[:, :, x], mags[:, :, x] = cart2pol(np.real(fftim1), np.imag(fftim1))

    if tarmag is None:
        tarmag = np.mean(mags, 2)

    xt, yt = tarmag.shape
    assert (xs == xt) and (ys == yt), 'The target spectrum must have the same size as the images.'
    f1 = np.linspace(-ys / 2, ys / 2 - 1, ys)
    f2 = np.linspace(-xs / 2, xs / 2 - 1, xs)
    XX, YY = np.meshgrid(f1, f2)
    t, r = cart2pol(XX, YY)
    if xs % 2 == 1 or ys % 2 == 1:
        r = np.round(r) - 1
    else:
        r = np.round(r)
    output_images = []
    for x in range(numin):
        fftim = mags[:, :, x]
        a = fftim.T.ravel()
        accmap = r.T.ravel() + 1
        a2 = tarmag.T.ravel()
        en_old = np.array(
            [np.sum([a[x] for x in y]) for y in [list(np.where(accmap == z)) for z in np.unique(accmap).tolist()]])
        en_new = np.array(
            [np.sum([a2[x] for x in y]) for y in [list(np.where(accmap == z)) for z in np.unique(accmap).tolist()]])
        coefficient = en_new / en_old
        cmat = coefficient[(r).astype(int)]  # coefficient[r+1]
        cmat[r > np.floor(np.max((xs, ys)) / 2)] = 0
        newmag = fftim * cmat
        XX, YY = pol2cart(angs[:, :, x], newmag)
        new = XX + YY * complex(0, 1)
        output = np.real(fft.ifft2(fft.ifftshift(new)))
        if rescaling == 0:
            output = (output * 255)
        output_images.append(output)
    if rescaling != 0:
        output_images = rescale_shine(output_images, rescaling)
    return output_images


def rescale_shine(images, option=1):
    assert type(images) == type([]), 'The input must be a list.'
    assert option == 1 or option == 2, "Invalid rescaling option"
    numin = len(images)
    brightests = np.zeros((numin, 1))
    darkests = np.zeros((numin, 1))
    for n in range(numin):
        if len(images[n].shape) == 3:
            images[n] = rgb2gray(images[n])
        brightests[n] = np.max(images[n])
        darkests[n] = np.min(images[n])
    the_brightest = np.max(brightests)
    the_darkest = np.min(darkests)
    avg_brightest = np.mean(brightests)
    avg_darkest = np.mean(darkests)
    output_images = []
    for m in range(numin):
        if option == 1:
            rescaled = (images[m] - the_darkest) / (the_brightest - the_darkest) * 255
        else:  # option == 2:
            rescaled = (images[m] - avg_darkest) / (avg_brightest - avg_darkest) * 255
        output_images.append(rescaled.astype(np.uint8))
    return output_images


def lumMatch(images, mask=None, lum=None):
    assert type(images) == type([]), 'The input must be a list.'
    assert (mask is None) or type(mask) == type([]), 'The input mask must be a list.'

    numin = len(images)
    if (mask is None) and (lum is None):
        M = 0;
        S = 0
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            M = M + np.mean(images[im])
            S = S + np.std(images[im])
        M = M / numin
        S = S / numin
        output_images = []
        for im in range(numin):
            im1 = copy.deepcopy(images[im])
            if np.std(im1) != 0:
                im1 = (im1 - np.mean(im1)) / np.std(im1) * S + M
            else:
                im1[:, :] = M
            output_images.append(im1)
    elif (mask is None) and (lum is not None):
        M = 0
        S = 0
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            M = lum[0]
            S = lum[1]
        M = M / numin
        S = S / numin
        output_images = []
        for im in range(numin):
            im1 = copy.deepcopy(images[im])
            if np.std(im1) != 0:
                im1 = (im1 - np.mean(im1)) / np.std(im1) * S + M
            else:
                im1[:, :] = M
            output_images.append(im1)
    elif (mask is not None) and (lum is None):
        M = 0
        S = 0
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            im1 = images[im]
            assert len(images) == len(mask), "The inputs must have the same length"
            m = mask[im]
            assert m.size == images[im].size, "Image and mask are not the same size"
            assert np.sum(m == 1) > 0, 'The mask must contain some ones.'
            M = M + np.mean(im1[m == 1])
            S = S + np.mean(im1[m == 1])
        M = M / numin
        S = S / numin
        output_images = []
        for im in range(numin):
            im1 = images[im]
            if type(mask) == type([]):
                m = mask[im]
            if np.std(im1[m == 1]):
                im1[m == 1] = (im1[m == 1] - np.mean(im1[m == 1])) / np.std(im1[m == 1]) * S + M
            else:
                im1[m == 1] = M
            output_images.append(im1)
    elif (mask is not None) and (lum is not None):
        M = lum[0]
        S = lum[1]
        output_images = []
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            im1 = images[im]
            if len(mask) == 0:
                if np.std(im1) != 0.0:
                    im1 = (im1 - np.mean(im1)) / np.std(im1) * S + M
                else:
                    im1[:, :] = M
            else:
                if type(mask) == type([]):
                    assert len(images) == len(mask), "The inputs must have the same length"
                    m = mask[im]
                assert m.size == images[im].size, "Image and mask are not the same size"
                assert np.sum(m == 1) > 0, 'The mask must contain some ones.'
                if np.std(im1[m == 1]) != 0.0:
                    im1[m == 1] = (im1[m == 1] - np.mean(im1[m == 1])) / np.std(im1[m == 1]) * S + M
                else:
                    im1[m == 1] = M
            output_images.append(im1)
    return output_images



# Configuration file for the Sphinx documentation builder.

# -- Project information
import os
import sys
from sphinx_automodapi import automodsumm
from sphinx_automodapi.utils import find_mod_objs
sys.path.insert(0, os.path.abspath('../..'))


project = 'SyMBac'
copyright = '2022, Georgeos Hardo'
author = 'Georgeos Hardo'

release = '0.2'
version = '0.2.0'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'matplotlib.sphinxext.plot_directive',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode'
]

autoapi_dirs = ['../../SyMBac']


intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

html_sidebars = { '**': ['globaltoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html'] }

# -- Options for EPUB output
epub_show_urls = 'footnote'

# Options for plot output
# -----------------------
plot_include_source = True

# numpydoc stuff
numpydoc_class_members_toctree = True
numpydoc_xref_param_type = True
numpydoc_validation_checks = {"all", "GL01", "SA04", "RT03"}
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"



import numpy as np
from scipy.ndimage import distance_transform_edt
from skimage.measure import label
import tifffile
import pkgutil
from io import BytesIO

from skimage.segmentation import find_boundaries
from skimage.transform import resize


def resize_mask(mask, resize_shape, ret_label):
    """
    Resize masks while maintaining their connectivity and values
    """
    labeled_mask = label(mask > 0, connectivity=1)
    labeled_mask = resize(labeled_mask, resize_shape, order=0, mode='reflect', cval=0, clip=True, preserve_range=True,
                          anti_aliasing=False, anti_aliasing_sigma=None).astype(int)
    mask_borders = find_boundaries(labeled_mask, mode="thick", connectivity=1)
    labeled_mask = np.where(mask_borders, 0, labeled_mask)
    if ret_label:
        return labeled_mask
    else:
        return labeled_mask > 0


def histogram_intersection(h1, h2, bins):
    sm = 0
    for i in range(bins):
        sm += min(h1[i], h2[i])
    return sm


def misc_load_img(dir):
    return tifffile.imread(BytesIO(pkgutil.get_data(__name__, dir)))


def get_sample_images():
    Ecoli100x = misc_load_img("sample_images/sample_100x.tiff")
    Ecoli100x_stationary = misc_load_img("sample_images/sample_100x_stationary.tiff")
    return {
        "E. coli 100x": Ecoli100x,
        "E. coli 100x stationary": Ecoli100x_stationary
    }


def unet_weight_map(y, wc=None, w0=10, sigma=5):
    """
    Generate weight maps as specified in the U-Net paper
    for boolean mask.

    "U-Net: Convolutional Networks for Biomedical Image Segmentation"
    https://arxiv.org/pdf/1505.04597.pdf

    Parameters
    ----------
    mask: Numpy array
        2D array of shape (image_height, image_width) representing binary mask
        of objects.
    wc: dict
        Dictionary of weight classes.
    w0: int
        Border weight parameter.
    sigma: int
        Border width parameter.

    Returns
    -------
    Numpy array
        Training weights. A 2D array of shape (image_height, image_width).
    """

    labels = label(y)
    no_labels = labels == 0
    label_ids = sorted(np.unique(labels))[1:]

    if len(label_ids) > 1:
        distances = np.zeros((y.shape[0], y.shape[1], len(label_ids)))

        for i, label_id in enumerate(label_ids):
            distances[:, :, i] = distance_transform_edt(labels != label_id)

        distances = np.sort(distances, axis=2)
        d1 = distances[:, :, 0]
        d2 = distances[:, :, 1]
        w = w0 * np.exp(-1 / 2 * ((d1 + d2) / sigma) ** 2) * no_labels
    else:
        w = np.zeros_like(y)
    if wc:
        class_weights = np.zeros_like(y)
        for k, v in wc.items():
            class_weights[y == k] = v
        w = w + class_weights
    return w



import numpy as np
from skimage.color import rgb2gray
import copy


# from https://stackoverflow.com/questions/20924085/python-conversion-between-coordinates
def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (phi, rho)


def pol2cart(phi, rho):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return (x, y)


def sfMatch(images, rescaling=0, tarmag=None):
    assert type(images) == type([]), 'The input must be a list.'

    numin = len(images)
    xs, ys = images[1].shape
    angs = np.zeros((xs, ys, numin))
    mags = np.zeros((xs, ys, numin))
    for x in range(numin):
        if len(images[x].shape) == 3:
            images[x] = rgb2gray(images[x])
        im1 = images[x] / 255
        xs1, ys1 = im1.shape
        assert (xs == xs1) and (ys == ys1), 'All images must have the same size.'
        fftim1 = np.fft.fftshift(np.fft.fft2(im1))
        angs[:, :, x], mags[:, :, x] = cart2pol(np.real(fftim1), np.imag(fftim1))

    if tarmag is None:
        tarmag = np.mean(mags, 2)

    xt, yt = tarmag.shape
    assert (xs == xt) and (ys == yt), 'The target spectrum must have the same size as the images.'
    f1 = np.linspace(-ys / 2, ys / 2 - 1, ys)
    f2 = np.linspace(-xs / 2, xs / 2 - 1, xs)
    XX, YY = np.meshgrid(f1, f2)
    t, r = cart2pol(XX, YY)
    if xs % 2 == 1 or ys % 2 == 1:
        r = np.round(r) - 1
    else:
        r = np.round(r)
    output_images = []
    for x in range(numin):
        fftim = mags[:, :, x]
        a = fftim.T.ravel()
        accmap = r.T.ravel() + 1
        a2 = tarmag.T.ravel()
        en_old = np.array(
            [np.sum([a[x] for x in y]) for y in [list(np.where(accmap == z)) for z in np.unique(accmap).tolist()]])
        en_new = np.array(
            [np.sum([a2[x] for x in y]) for y in [list(np.where(accmap == z)) for z in np.unique(accmap).tolist()]])
        coefficient = en_new / en_old
        cmat = coefficient[(r).astype(int)]  # coefficient[r+1]
        cmat[r > np.floor(np.max((xs, ys)) / 2)] = 0
        newmag = fftim * cmat
        XX, YY = pol2cart(angs[:, :, x], newmag)
        new = XX + YY * complex(0, 1)
        output = np.real(np.fft.ifft2(np.fft.ifftshift(new)))
        if rescaling == 0:
            output = (output * 255).astype(np.uint8)
        output_images.append(output)
    if rescaling != 0:
        output_images = rescale(output_images, rescaling)
    return output_images


def rescale(images, option=1):
    assert type(images) == type([]), 'The input must be a list.'
    assert option == 1 or option == 2, "Invalid rescaling option"
    numin = len(images)
    brightests = np.zeros((numin, 1))
    darkests = np.zeros((numin, 1))
    for n in range(numin):
        if len(images[n].shape) == 3:
            images[n] = rgb2gray(images[n])
        brightests[n] = np.max(images[n])
        darkests[n] = np.min(images[n])
    the_brightest = np.max(brightests)
    the_darkest = np.min(darkests)
    avg_brightest = np.mean(brightests)
    avg_darkest = np.mean(darkests)
    output_images = []
    for m in range(numin):
        if option == 1:
            rescaled = (images[m] - the_darkest) / (the_brightest - the_darkest) * 255
        elif option == 2:
            rescaled = (images[m] - avg_darkest) / (avg_brightest - avg_darkest) * 255
        output_images.append(rescaled.astype(np.uint8))
    return output_images


def lumMatch(images, mask=None, lum=None):
    assert type(images) == type([]), 'The input must be a list.'
    assert (mask is None) or type(mask) == type([]), 'The input mask must be a list.'

    numin = len(images)
    if (mask is None) and (lum is None):
        print("HI")
        M = 0
        S = 0
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            M = M + np.mean(images[im])
            S = S + np.std(images[im])
        M = M / numin
        S = S / numin
        output_images = []
        for im in range(numin):
            im1 = copy.deepcopy(images[im])
            if np.std(im1) != 0:
                print("did_std")
                im1 = (im1 - np.mean(im1)) / np.std(im1) * S + M
            else:
                im1[:, :] = M
            output_images.append(im1.astype(np.uint8))
    elif (mask is not None) and (lum is None):
        print("HI2")
        M = 0
        S = 0
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            im1 = images[im]
            assert len(images) == len(mask), "The inputs must have the same length"
            m = mask[im]
            assert m.size == images[im].size, "Image and mask are not the same size"
            assert np.sum(m == 1) > 0, 'The mask must contain some ones.'
            M = M + np.mean(im1[m == 1])
            S = S + np.mean(im1[m == 1])
        M = M / numin
        S = S / numin
        output_images = []
        for im in range(numin):
            im1 = images[im]
            if type(mask) == type([]):
                m = mask[im]
            if np.std(im1[m == 1]):
                im1[m == 1] = (im1[m == 1] - np.mean(im1[m == 1])) / np.std(im1[m == 1]) * S + M
            else:
                im1[m == 1] = M
            output_images.append(im1.astype(np.uint8))
    elif (mask is not None) and (lum is not None):
        print("HI3")
        M = lum[0]
        S = lum[1]
        output_images = []
        for im in range(numin):
            if len(images[im].shape) == 3:
                images[im] = rgb2gray(images[im])
            im1 = images[im]
            if len(mask) == 0:
                if np.std(im1) != 0.0:
                    im1 = (im1 - np.mean(im1)) / np.std(im1) * S + M
                else:
                    im1[:, :] = M
            else:
                if type(mask) == type([]):
                    assert len(images) == len(mask), "The inputs must have the same length"
                    m = mask[im]
                assert m.size == images[im].size, "Image and mask are not the same size"
                assert np.sum(m == 1) > 0, 'The mask must contain some ones.'
                if np.std(im1[m == 1]) != 0.0:
                    im1[m == 1] = (im1[m == 1] - np.mean(im1[m == 1])) / np.std(im1[m == 1]) * S + M
                else:
                    im1[m == 1] = M
            output_images.append(im1.astype(np.uint8))
    return output_images



from setuptools import setup
from pathlib import Path
this_directory = Path(__file__).parent

setup(
    name = 'SyMBac',
    version = '0.2.0',    
    description = 'A package for generating synthetic images of bactera in phase contrast or fluorescence. Used for creating training data for machine learning segmentation and tracking algorithms.',
    url = 'https://github.com/georgeoshardo/SyMBac',
    author = 'Georgeos Hardo',
    author_email = 'gh464@cam.ac.uk',
    license = 'GPL-2.0',
    packages = ['SyMBac', 'SyMBac.external', 'SyMBac.sample_images', 'SyMBac.external.DeLTA'],
    package_data = {'': ['sample_images/*.tiff']},
    include_package_data = True,
    long_description = (this_directory / "README.md").read_text(), #"SyMBac is a tool to generate synthetic phase
    # contrast or fluorescence images of bacteria. Currently the tool only supports bacteria growing in the mother
    # machine, however support for bacteria growing in monolayers (and maybe even biofilms!) is coming.",
    long_description_content_type = 'text/markdown',
    install_requires = [
        'tensorflow==2.8.0',
        'elasticdeform==0.4.9', 
        'tifffile==2021.10.12',
        'scikit-image==0.18.3' ,
        'matplotlib==3.4.3',
        "tqdm==4.62.3",
        "pandas==1.3.4",
        "natsort==7.1.1",
        "jupyterlab",
        "ipywidgets",
        "joblib==1.1.0",
        "napari[all]",
        "pymunk==6.2.0",
        "pyglet==1.5.21",
        "raster-geometry==0.1.4.1",
        "matplotlib-scalebar==0.7.2"
        ],

    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',  
        'Operating System :: POSIX :: Linux', 
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering'
    ],
    test_suite = "tests.test_phase_contrast_drawing",
)



import numpy as np
def circ(theta, start, radius):
    y = radius * np.cos(theta) +radius
    x = radius * np.sin(theta) + start + radius
    return x, y

def wall(thickness, start, end, t_or_b, resolution):
    """Generates the straight part of the cell wall's coordinates (all but the poles)

    Parameters
    ----------
    thickness : float
        The distance from the top cell wall to the bottom cell wall
    start : float
        The start coordinate of the cell wall
    end : float
        The end coordinate of the cell wall
    t_or_b: int
        0 for top wall
        1 for bottom wall
    resolution : int
        Number of points defining the cell wall geometry
    Returns
    -------
    2-tuple of 1D numpy arrays
        return[0] is the wall's x coordinates
        return[0] is the wall's y coordiantes
        
    Example
    -------
    Create two cell walls of length 10, 3 apart
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> top_wall = wall(3,0,10,0,20)
    >>> bottom_wall = top_wall = wall(3,0,10,1,20)
    >>> plt.plot(walls[0], walls[1])
    >>> plt.plot(walls[0], walls[1])
    >>> plt.show()
    """
    wall_x = np.linspace(start, end, resolution)
    wall_y = np.ones(resolution)*thickness * t_or_b +thickness
    return wall_x, wall_y

def get_vertices(cell_length, cell_width, angle, resolution):
    """Generates coordinates for a cell centered aronud (0,0)

    Parameters
    ----------
    cell_length : float
        The length of the STRAIGHT part of the cell's wall. 
        Total length is cell_length + cell_width because the poles are models as semi-circles.
    cell_width : float
        Total thickness of the cell, defines the poles too.
    angle : float
        Angle in radians to rotate the cell by (counter-clockwise)
    resolution : int
        Number of points defining the cell wall geometry
    Returns
    -------
    list of lists containing cell x and y coords
    
    Example
    -------
    Create a cell of length 10+4 rotated by 1 radian with a resolution of 20:
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> verts = get_vertices(10,4,1,20)
    >>> verts_y = [y[0] for y in verts]
    >>> verts_x = [x[1] for x in verts]
    >>> plt.plot(verts_x,verts_y)
    """
    cell_width = cell_width/2
    left_wall = circ(np.linspace(np.pi,2*np.pi, resolution), 0, cell_width)
    top_wall_xy = wall(cell_width, cell_width, cell_length, 1, resolution)
    bottom_wall_xy = wall(cell_width, cell_width, cell_length, -1, resolution)
    right_wall = circ(np.linspace(0,np.pi, resolution), cell_length - cell_width, cell_width)
    coordinates = [[left_wall[0][x] - cell_length/2, left_wall[1][x] - cell_width/2] for x in reversed(range(len(left_wall[0])))] + \
            [[bottom_wall_xy[0][x] - cell_length/2, bottom_wall_xy[1][x]- cell_width/2] for x in (range(len(bottom_wall_xy[0])))] + \
            [[right_wall[0][x] - cell_length/2, right_wall[1][x]- cell_width/2] for x in reversed(range(len(right_wall[0])))] + \
            [[top_wall_xy[0][x] - cell_length/2, top_wall_xy[1][x]- cell_width/2] for x in reversed(range(len(top_wall_xy[0])))]
    coordinates = np.array(coordinates)
    cell_centroid = centroid(coordinates)
    centered_verts = coordinates - cell_centroid
    centered_verts = centered_verts.tolist()

    rotated = np.zeros((len(centered_verts),2))
    for x in range(len(centered_verts)):
        rotated[x] = rotate(cell_centroid, (centered_verts[x][0],centered_verts[x][1]), angle)
    centered_verts = rotated - centroid(rotated)

    return centered_verts.tolist()

def centroid(vertices):
    """Return the centroid of a list of vertices 
    
    Keyword arguments:
    vertices -- A list of tuples containing x,y coordinates.

    """
    return np.sum(vertices,axis=0)/len(vertices)

def rotate(origin, point, angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.

    The angle should be given in radians.
    """
    ox, oy = origin
    px, py = point

    qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
    qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
    return qx, qy



import numpy as np
import pymunk
from SyMBac import cell_geometry
class Cell:
    """
    A class defining the cell objects
    
    ...
    
    Attributes
    ----------
    length : float
        Cell's length
    width : float
        Cell's width
    resolution : int
        Number of points defining cell's geometry
    position : (float, float)
        x,y coords of cell centroid
    angle : float
        rotation in radians of cell (counterclockwise)
    space : pymunk.space.Space
        The pymunk space of the cell
    dt : float
        Timestep the cell experiences every iteration
    growth_rate_constant : float
        The cell grows by a function of dt*growth_rate_constant depending on its growth model
    max_length : float
        The maximum length a cell reaches before dividing
    max_length_mean : float
        should be the same as max_length for reasons unless doing advanced simulations
    max_length_var : float
        The variance defining a normal distribution around max_length
    width_var : float
        The variance defining a normal distribution around width
    width_mean : float
        For reasons should be set equal to width unless using advanced features
    """
    def __init__(
        self,
        length,
        width,
        resolution,
        position,
        angle,
        space,
        dt,
        growth_rate_constant,
        max_length,
        max_length_mean,
        max_length_var,
        width_var,
        width_mean,
        parent = None,
        daughter = None, 
    ):
        
        """
        Initialising a cell

        Parameters
        ----------
        length : float
            Cell's length
        width : float
            Cell's width
        resolution : int
            Number of points defining cell's geometry
        position : (float, float)
            x,y coords of cell centroid
        angle : float
            rotation in radians of cell (counterclockwise)
        space : pymunk.space.Space
            The pymunk space of the cell
        dt : float
            Timestep the cell experiences every iteration
        growth_rate_constant : float
            The cell grows by a function of dt*growth_rate_constant depending on its growth model
        max_length : float
            The maximum length a cell reaches before dividing
        max_length_mean : float
            should be the same as max_length for reasons unless doing advanced simulations
        max_length_var : float
            The variance defining a normal distribution around max_length
        width_var : float
            The variance defining a normal distribution around width
        width_mean : float
            For reasons should be set equal to width unless using advanced features
        body : pymunk.body.Body
            The cell's pymunk body object
        shape : pymunk.shapes.Poly
            The cell's pymunk body object
        ID : int
            A unique identifier for each cell. At the moment just a number from 0 to 100_000_000 and cross fingers that we get no collisions. 
            
        For info about the Pymunk objects, see the API reference. http://www.pymunk.org/en/latest/pymunk.html
        Cell class has been tested and works with pymunk version 6.0.0

        Returns
        -------
        outupt : None
        """
        self.dt = dt
        self.growth_rate_constant = growth_rate_constant
        self.length = length
        self.width_mean = width_mean
        self.width_var = width_var
        self.width = width
        self.resolution = resolution
        self.angle = angle
        self.position = position
        self.space = space
        self.max_length = max_length
        self.max_length_mean = max_length_mean
        self.max_length_var = max_length_var
        self.body, self.shape = self.create_pm_cell()
        self.angle = self.body.angle
        self.ID = np.random.randint(0,100_000_000)
        self.parent = parent
        self.daughter = daughter     
        
        
        

    def create_pm_cell(self):
        if self.is_dividing():
            new_length = self.length/2
            daughter_length = self.length - new_length
            self.length = new_length
            cell_vertices = self.calculate_vertex_list()
            cell_shape = pymunk.Poly(None, cell_vertices)
            self.shape = cell_shape
            cell_moment = 1
            cell_mass = 0.0001
            cell_body = pymunk.Body(cell_mass,cell_moment)
            cell_shape.body = cell_body
            self.body = cell_body
            new_x = self.position[0] + self.length/2 * np.cos(self.angle*2)
            new_y = self.position[1] + self.length/2 * np.sin(self.angle*2)
            self.body.position = [new_x, new_y]
            cell_body.angle = self.angle
            cell_shape.friction=1
            self.space.add(cell_body, cell_shape)
            daughter_details = {
                "length": daughter_length*0.9,
                "width": np.random.normal(self.width_mean,self.width_var),
                "resolution": self.resolution,
                "position": [self.position[0] - self.length/2 * np.cos(self.angle*2), self.position[1] - self.length/2 * np.sin(self.angle*2)],
                "angle": self.angle*np.random.uniform(0.95,1.05),
                "space": self.space,
                "dt": self.dt,
                "growth_rate_constant": self.growth_rate_constant,
                "max_length": np.random.normal(self.max_length_mean,self.max_length_var),
                "max_length_mean": self.max_length_mean,
                "max_length_var": self.max_length_var,
                "width_var": self.width_var,
                "width_mean": self.width_mean,
                "parent": self
            }
            return daughter_details
        else:
            cell_vertices = self.calculate_vertex_list()
            cell_shape = pymunk.Poly(None, cell_vertices)
            self.shape = cell_shape
            cell_moment = 1
            cell_mass = 0.0001
            cell_body = pymunk.Body(cell_mass,cell_moment)
            cell_shape.body = cell_body
            self.body = cell_body
            cell_body.position = self.position
            cell_body.angle = self.angle
            cell_shape.friction=0.01
            self.space.add(cell_body, cell_shape)
            return cell_body, cell_shape

    def is_dividing(self): # This needs to be made constant or a cell can divide in one frame and not another frame
        if self.length > (self.max_length):
            return True
        else:
            return False


    def update_length(self):
        self.length = self.length + self.growth_rate_constant*self.dt*self.length*np.random.uniform(0.7,1.3)

    def update_position(self):
        self.position = self.body.position
        self.angle = self.body.angle

    def get_angle(self):
        return self.body.angle
    
    def calculate_vertex_list(self):
        return cell_geometry.get_vertices(
            self.length,
            self.width,
            self.angle, 
            self.resolution
            )

    def get_vertex_list(self):
        vertices = []
        for v in self.shape.get_vertices():
            x,y = v.rotated(self.shape.body.angle) + self.shape.body.position #.rotated(self.shape.body.angle)
            vertices.append((x,y))
        return vertices

    def get_centroid(self):
        vertices = self.get_vertex_list()
        return cell_geometry.centroid(vertices)

    


import numpy as np
import pymunk
from SyMBac import cell_geometry

class Cell:
    def __init__(
        self,
        length,
        width,
        resolution,
        position,
        angle,
        space,
        dt,
        growth_rate_constant,
        max_length,
        max_length_mean,
        max_length_var,
        width_var,
        width_mean
    ):
        self.dt = dt
        self.growth_rate_constant = growth_rate_constant
        self.length = length
        self.width_mean = width_mean
        self.width_var = width_var
        self.width = width
        self.resolution = resolution
        self.angle = angle
        self.position = position
        self.space = space
        self.max_length = max_length
        self.max_length_mean = max_length_mean
        self.max_length_var = max_length_var
        print(self.create_pm_cell())
        self.body, self.shape = self.create_pm_cell()
        self.angle = self.body.angle
        self.ID = np.random.randint(0,100_000_000)
        
        
        
        

    def create_pm_cell(self):
        if self.is_dividing() == True:
            new_length = self.length/2
            daughter_length = self.length - new_length
            self.length = new_length
            cell_vertices = self.calculate_vertex_list()
            cell_shape = pymunk.Poly(None, cell_vertices)
            self.shape = cell_shape
            cell_moment = 1
            cell_mass = 0.0001
            cell_body = pymunk.Body(cell_mass,cell_moment)
            cell_shape.body = cell_body
            self.body = cell_body
            new_x = self.position[0] + self.length/2 * np.cos(self.angle*2)
            new_y = self.position[1] + self.length/2 * np.sin(self.angle*2)
            self.body.position = [new_x, new_y]
            cell_body.angle = self.angle
            cell_shape.friction=1000
            self.space.add(cell_body, cell_shape)
            daughter_details = {
                "length": daughter_length,
                "width": np.random.normal(self.width_mean,self.width_var),
                "resolution": self.resolution,
                "position": [self.position[0] - self.length/2 * np.cos(self.angle*2), self.position[1] - self.length/2 * np.sin(self.angle*2)],
                "angle": self.angle*np.random.uniform(0.85,1.15), # CHANGE THIS BACK FOR MM
                "space": self.space,
                "dt": self.dt,
                "growth_rate_constant": self.growth_rate_constant,
                "max_length": np.random.normal(self.max_length_mean,self.max_length_var),
                "max_length_mean": self.max_length_mean,
                "max_length_var": self.max_length_var,
                "width_var": self.width_var,
                "width_mean": self.width_mean
            }
            return daughter_details
        else:
            cell_vertices = self.calculate_vertex_list()
            cell_shape = pymunk.Poly(None, cell_vertices)
            self.shape = cell_shape
            cell_moment = 1
            cell_mass = 0.0001
            cell_body = pymunk.Body(cell_mass,cell_moment)
            cell_shape.body = cell_body
            self.body = cell_body
            cell_body.position = self.position
            cell_body.angle = self.angle
            cell_shape.friction=1000
            self.space.add(cell_body, cell_shape)
            return cell_body, cell_shape

    def is_dividing(self): # This needs to be made constant or a cell can divide in one frame and not another frame
        if self.length > (self.max_length):
            return True
        else:
            return False

    def growth_modifier(self, xpos, ypos):
        return 2 * (1/(1 + np.exp(-0.005 * abs(xpos-400))) - 0.5) * (1/(1 + np.exp(-0.1 * abs(ypos-400))) - 0.5) + 1/(len(self.space.bodies)+2)
    #This was changed for the colony
    
    def update_length(self):
        self.length = self.length + self.growth_rate_constant*self.dt*self.length*np.random.uniform(0.5,1.2)* self.growth_modifier(self.position[0], self.position[1])
        for _ in range(50+len(self.space.bodies)):
            self.space.step(self.dt)
    
    def update_position(self):
        self.position = self.body.position
        self.angle = self.body.angle
    
    def get_angle(self):
        return self.body.angle
    
    def calculate_vertex_list(self):
        return cell_geometry.get_vertices(
            self.length,
            self.width,
            self.angle, 
            self.resolution
            )

    def get_vertex_list(self):
        vertices = []
        for v in self.shape.get_vertices():
            x,y = v.rotated(self.shape.body.angle) + self.shape.body.position #.rotated(self.shape.body.angle)
            vertices.append((x,y))
        return vertices

    def get_centroid(self):
        vertices = self.get_vertex_list()
        return cell_geometry.centroid(vertices)

    


"""
This script process an input file / folder with data from a multi-position 
microscopy experiment and produces .mat files describing cells lineage, 
fluorescence etc.

The script can be called from the command line:
    python pipeline.py /path/to/file/or/folder /path/to/output/folder
    
Or run from an IDE by changing file paths.

@author: jblugagne
"""

import config as cfg
# Modules:
import time, cv2, os, sys
import numpy as np
from data import postprocess
import utilities as utils


# Load models:
models = utils.loadmodels()

def process_xp(xpfile,
               positions=None,
               use_bioformats=False,
               prototype=None, 
               fileorder='pct',
               filenamesindexing=1,
               resfolder = None,
               rotation_correction = True,
               write_movie = True,
               verbose = 1):
    '''
    Process entire experiment folder.

    Parameters
    ----------
    xpfile : str or tuple/list of str
        Path to experiment file(s) or folder.
    positions : list or None, optional
        List of positions to process. If multiple files, list of lists of 
        positions. If None, all positions are processed.
        The default is None.
    use_bioformats : bool, optional
        Use bio-formats reader for provided files (necessary for .nd2, .oib, 
        .czi...)
        The default is False.
    prototype : str, optional
        File names prototype. See xpreader() in utilities.py.
        The default is None.
    fileorder : str, optional
        File names order. See xpreader() in utilities.py. 
        The default is 'pct'.
    filenamesindexing : int, optional
        File names base indexing. See xpreader() in utilities.py.
        The default is 1.
    resfolder : str, optional
        Path to save results to. If None, a folder within the experiment folder
        will be created.
        The default is None.
    rotation_correction : bool, optional
        Flag to automatically correct the angle of images. 
        The default is True.
    write_movie : bool, optional
        Flag to write movies of results to disk.
        The default is True.
    verbose: int, optional
        Verbosity of console output.
        The default is 1.

    Returns
    -------
    None.

    '''
    
    # If single file path str, turn into list:    
    if type(xpfile) is str:
        xpfile = [xpfile]
    
    # Loop through experiments:
    for x, experiment in enumerate(xpfile):
        
        # Print start message:
        if verbose >= 1:
            print('\n%s ----- Processing experiment %d/%d: %s'%(time.strftime("%e %b %Y - %T", 
                                                                            time.localtime()),
                                                              x+1,
                                                              len(xpfile),
                                                              os.path.split(experiment)[1])
                  )
        
        # Create folders if they don't exist:
        if resfolder is None:
            if os.path.isdir(experiment):
                resfolder = os.path.join(experiment, 'delta_results')
            else:
                resfolder = os.path.splitext(experiment)[0]+'_delta_results'
        if not os.path.exists(resfolder):
            os.mkdir(resfolder)
        
        # Initialize XP reader:
        xp = utils.xpreader(experiment,
                            prototype=prototype,
                            fileorder=fileorder,
                            filenamesindexing=filenamesindexing,
                            use_bioformats=use_bioformats)
        
        # Process positions to run:
        if positions is None:
            positions_torun = range(xp.positions)
        elif len(xpfile)==1:
            positions_torun = positions
        else:
            positions_torun = positions[x]
        if positions_torun is None:
            positions_torun = range(xp.positions)
        
        # Run over all positions:
        for pos in positions_torun:
            
            # Run processing for position:
            delta_output = process_position(xp,
                                            pos,
                                            verbose=verbose)
    
            # Save results to disk:
            utils.legacysave(os.path.join(resfolder,('Position%06d.mat' % (pos+1))), 
                             experiment, 
                             *delta_output,
                             [xp.y,xp.x,xp.channels,xp.timepoints])
            
            # Write results movie:
            if write_movie:
                if verbose>=1:
                    print('Writing mp4 movie file...')
                movie = utils.results_movie(xp,
                                            pos,
                                            *delta_output,
                                            display=False
                                            )
                utils.vidwrite(movie*255,
                               os.path.join(resfolder,('Position%06d.mp4' % (pos+1))),
                               verbose=0
                               )
            
            
            
    # Done, print message:
    if verbose>=1:
        print('\n\n%s ----- %d/%d experiments processed'%(time.strftime("%e %b %Y - %T", 
                                                                        time.localtime()),
                                                          x+1,
                                                          len(xpfile))
              )
        
def process_position(xp,
                     pos,
                     rotation_correction = True,
                     verbose = 1):
    '''
    Apply DeLTA pipeline to a single position in the experiment

    Parameters
    ----------
    xp : xpreader object
        See xpreader() in utilities.py. 
    pos : int
        Position to process.
    rotation_correction : bool, optional
        Flag to automatically correct the angle of images. 
        The default is True.
    verbose: int, optional
        Verbosity of console output.
        The default is 1.

    Returns
    -------
    res : list
        List of dictionaries containing the lineage and labels stacks.
    chamberboxes : list
        List of dictionaries containing the positions of chamber crop boxes in
        the image.
    rot_corr : float
        Rotation applied to images to get the chambers horizontal.
    drift_values : list
        XY drift values computed by preprocessing step.

    '''
    
    ##### Preprocessing - Rotation correction, Chambers identificaton, Drift:
    start_time = time.time()
    if verbose: print('Pos. %d/%d: Preprocessing...' %(pos+1,xp.positions), end='')
    transframes, fluoframes, chamberboxes, rot_corr, driftvalues = preprocessing(xp, pos)
    if chamberboxes is None:
        if verbose: print('\b\b\b - No chambers detected! Aborting...')
        return None, None, rot_corr, None
    
    
    ##### Cell segmentation:
    if verbose: print('\b\b\b - Segmentation...', end='')
    seg, seg_inputs = segmentation(transframes, chamberboxes)
    
    
    ##### Tracking:
    res = trackinglineagefeatures(transframes, 
                                  fluoframes, 
                                  seg, 
                                  seg_inputs,
                                  chamberboxes)
    
    if verbose: print(' - Done! ['+str(round(time.time() - start_time))+' secs]')
    
    return res, chamberboxes, rot_corr, driftvalues

def preprocessing(xp, pos, rotation_correction = True):
    '''
    Rotation correction, drift correction, loading images...

    Parameters
    ----------
    xp : xpreader object
        See xpreader() in utilities.py. 
    pos : int
        Position to process.
    rotation_correction : bool, optional
        Flag to automatically correct the angle of images. 
        The default is True.

    Returns
    -------
    transframes : 3D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- image_size_x -by- image_size_y.
    fluoframes : 4D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- fluo_channels -by- image_size_x -by- image_size_y.
    chamberboxes : list
        List of dictionaries containing the positions of chamber crop boxes in
        the image.
    rot_corr : float
        Rotation applied to images to get the chambers horizontal.
    drift_values : list
        XY drift values computed by preprocessing step.

    '''
    
    firstframe = xp.getframes(positions=pos,channels=0,frames=0,rescale=(0,1)) # Get first frame for rotation estimation (We assume first channel is trans)
    # Rotation correction:
    if rotation_correction:
        rot_corr = utils.deskew(firstframe) # Estimate rotation angle
    else:
        rot_corr = 0
    # Get trans frames:
    transframes = xp.getframes(positions=pos,channels=0,rescale=(0,1),rotate=rot_corr) # Now load all trans frames, with rotation correction
    firstframe = np.expand_dims(np.expand_dims(cv2.resize(transframes[0],(512,512)),axis=0),axis=3) # using expand_dims to get it into a shape that the chambers id unet accepts
    # Find chambers, filter results, get bounding boxes:
    chambermask = models['chambers'].predict(firstframe,verbose=0)
    chambermask = cv2.resize(np.squeeze(chambermask),transframes.shape[3:0:-1]) # scaling back to original size
    chambermask = postprocess(chambermask,min_size=cfg.min_chamber_area) # Binarization, cleaning and area filtering
    chamberboxes = utils.getChamberBoxes(np.squeeze(chambermask))
    # If no chambers detected:
    if len(chamberboxes)==0:
        return None, None, None, rot_corr, None
        
    # Drift correction:
    drifttemplate = utils.getDriftTemplate(
        chamberboxes, transframes[0], whole_frame=cfg.whole_frame_drift
        ) # This template will be used as reference for drift correction
    driftcorbox = dict(
        xtl = 0,
        xbr = None,
        ytl = 0,
        ybr = None if cfg.whole_frame_drift else max(chamberboxes,
                                                     key=lambda elem: elem['ytl']
                                                     )['ytl']
        ) # Box to match template
    transframes, driftvalues = utils.driftcorr(transframes, template=drifttemplate, box=driftcorbox) # Run drift corr    
    # Load up fluoresence images, apply drift correction and rotation:
    fluoframes = xp.getframes(positions=pos, channels=[x for x in range(1,xp.channels)], squeeze_dimensions=False, rotate=rot_corr)
    if fluoframes is not None:
        fluoframes = fluoframes[0] # Squeezing out the position dimension
        for f in range(fluoframes.shape[1]): #Go over all channels:
            fluoframes[:,f,:,:], _ = utils.driftcorr(fluoframes[:,f,:,:], drift=driftvalues) # Apply drift correction
    
    return transframes, fluoframes, chamberboxes, rot_corr, driftvalues

def segmentation(transframes, chamberboxes):
    '''
    

    Parameters
    ----------
    transframes : 3D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- image_size_x -by- image_size_y.
    chamberboxes : list
        List of dictionaries containing the positions of chamber crop boxes in
        the image.

    Returns
    -------
    seg : 3D array of floats
        Segmentation Unet outputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    seg_inputs : 3D array of floats
        Segmentation Unet inputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.

    '''
    
    seg_inputs = []
    # Compile segmentation inputs:
    for m, chamberbox in enumerate(chamberboxes):
        for i in range(transframes.shape[0]):
            seg_inputs.append(cv2.resize(utils.rangescale(utils.cropbox(transframes[i],chamberbox),(0,1)),(32,256)))
    seg_inputs = np.expand_dims(np.array(seg_inputs),axis=3) # Format into 4D tensor
    # Run segmentation U-Net:
    seg = models['segmentation'].predict(seg_inputs,verbose=0)
    seg = postprocess(seg[:,:,:,0], min_size=cfg.min_cell_area)
    
    return seg, seg_inputs

def trackinglineagefeatures(transframes, 
                        fluoframes, 
                        seg, 
                        seg_inputs, 
                        chamberboxes,
                        verbose = 1):
    '''
    For each chamber, perform tracking, lineage reconstruction, and feature
    extraction.

    Parameters
    ----------
    transframes : 3D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- image_size_x -by- image_size_y.
    fluoframes : 4D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- fluo_channels -by- image_size_x -by- image_size_y.
    seg : 3D array of floats
        Segmentation Unet outputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    seg_inputs : 3D array of floats
        Segmentation Unet inputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    chamberboxes : list
        List of dictionaries containing the positions of chamber crop boxes in
        the image.
    verbose: int, optional
        Verbosity of console output.
        The default is 1.

    Returns
    -------
    res : list
        List of dictionaries containing the lineage and labels stacks.

    '''
    
    if verbose: 
        print('\b\b\b - Tracking & Feature extraction ', end='')
        message = ''
    res = []
    for m, chamberbox in enumerate(chamberboxes):
        if verbose: 
            print('\b'*len(message),end='')
            message = 'chamber '+str(m+1)+'/'+str(len(chamberboxes))
            print(message,end='')
        
        
        #### Tracking Unet:
        track, track_inputs, frame_numbers = tracking(seg, 
                                                      seg_inputs, 
                                                      m, 
                                                      transframes.shape[0])
        
        ##### Compile lineage and label stack:
        lin, label_stack = lineage(seg, 
                                   track, 
                                   track_inputs, 
                                   frame_numbers, 
                                   m, 
                                   transframes.shape[0])
            
        ##### Extract features:
        lin, label_stack_resized = features(lin,
                                            label_stack,
                                            fluoframes,
                                            chamberbox,
                                            transframes.shape[0])
        
            
        # Append to res dictionary:
        res.append(dict(lineage=lin,labelsstack=label_stack,labelsstack_resized=label_stack_resized))
    
    return res
    
def tracking(seg, seg_inputs, chamber_number, timepoints):
    '''
    Perform tracking for a single chamber.

    Parameters
    ----------
    seg : 3D array of floats
        Segmentation Unet outputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    seg_inputs : 3D array of floats
        Segmentation Unet inputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    chamber_number : int
        Number of the current chamber.
    timepoints : int
        Number of timepoints in movie.

    Returns
    -------
    track : 4D array of floats.
        Tracking Unet outputs. Dimensions are tracking events -by- 256 -by- 32
        -by- 3.
    track_inputs : 4D array of floats.
        Tracking Unet outputs. Dimensions are tracking events -by- 256 -by- 32
        -by- 4.
    frame_numbers : list
        Frame number / timepoint of each tracking event.

    '''
    
    track_inputs = [] # Here we work chamber by chamber for memory reasons
    frame_numbers = []# To reconstruct
    # Compile tracking inputs:
    for i in range(1,timepoints):
        singleseg = utils.getSinglecells(seg[chamber_number*timepoints + i-1]) # Get single cell masks in the 'old' frame
        for c in range(singleseg.shape[0]):
            frame_numbers.append(i) # For lineage reconstruction later
            track_inputs.append(np.stack((
                    seg_inputs[chamber_number*timepoints + i,:,:,0], # Current trans image
                    singleseg[c], # Mask of one previous cell
                    seg_inputs[chamber_number*timepoints + i-1,:,:,0], # Previous trans image
                    seg[chamber_number*timepoints + i]), # Mask of all current cells
                    axis=-1
                ))
    track_inputs = np.array(track_inputs) # Format into 4D tensor
    # Run U-Net model for tracking:
    if track_inputs.size!=0: # If not empty
        track = models['tracking'].predict(track_inputs,verbose=0)
    else:
        track = None
    
    return track, track_inputs, frame_numbers

def lineage(seg, track, track_inputs, frame_numbers, chamber_number, timepoints):
    '''
    Assemble lineage from tracking outputs.

    Parameters
    ----------
    seg : 3D array of floats
        Segmentation Unet outputs. Dimensions are chambers*timepoints -by- 
        256 -by- 32.
    track : 4D array of floats.
        Tracking Unet outputs. Dimensions are tracking events -by- 256 -by- 32
        -by- 3.
    track_inputs : 4D array of floats.
        Tracking Unet outputs. Dimensions are tracking events -by- 256 -by- 32
        -by- 4.
    frame_numbers : list
        Frame number / timepoint of each tracking event.
    chamber_number : int
        Number of the current chamber.
    timepoints : int
        Number of timepoints in movie.

    Returns
    -------
    lin : list
        List of dictionaries containing the information relative to each cell
        in a particular chamber.
    label_stack : 3D array of uint16
        Images stack labelled with each cell in the chamber, over time. For 
        each frame in the stack, i-valued pixels correspond to the (i-1)-th 
        list element/cell in the lin output (0-based list indexing). The 
        dimensions are timepoints -by- 256 -by- 32.

    '''
    
    label_stack = np.zeros([timepoints,seg.shape[1],seg.shape[2]],dtype=np.uint16)
    lin, label_stack = utils.updatelineage(seg[chamber_number*timepoints], label_stack) # Initialize lineage and label stack on first frame
    for i in range(1,timepoints):
        frame_idxs = [x for x, fn in enumerate(frame_numbers) if fn==i]
        if frame_idxs:
            scores = utils.getTrackingScores(track_inputs[frame_idxs[0],:,:,3], track[frame_idxs])
            attrib = utils.getAttributions(scores)
        else:
            attrib = []
        lin, label_stack = utils.updatelineage(seg[chamber_number*timepoints + i], label_stack, framenb=i, lineage=lin, attrib=attrib) # Because we use uint16, we can only track up to 65535 cells per chamber
    
    return lin, label_stack

def features(lin, label_stack, fluoframes, chamberbox, timepoints):
    '''
    Extract cell features such as fluorescence, cell length etc...

    Parameters
    ----------
    lin : list
        List of dictionaries containing the information relative to each cell
        in a particular chamber.
    label_stack : 3D array of uint16
        Images stack labelled with each cell in the chamber, over time. For 
        each frame in the stack, i-valued pixels correspond to the (i-1)-th 
        list element/cell in the lin output (0-based list indexing).
    fluoframes : 4D array of floats
        Transmitted light images, rescaled to [0, 1] dynamic range. Dimensions
        are timepoints -by- fluo_channels -by- image_size_x -by- image_size_y.
    chamberboxes : list
        Dictionary containing the position of the chamber crop box in the image
        for the current chamber.
    timepoints : int
        Number of timepoints in movie.

    Returns
    -------
    lin : list
        List of dictionaries containing the information relative to each cell
        in a particular chamber (updated with single-cell features).
    label_stack_resized : 3D array of uint16.
        Resized version of label_stack, scaled back to the original chamber 
        images crop box size.

    '''
    
    # Add dictionary keys for the features:
    for l in range(len(lin)):
        lin[l] = {**lin[l], **dict(length=[],width=[],area=[],pixels=[])}
        if fluoframes is not None:
            for f in range(fluoframes.shape[1]): #Go over all channels:
                lin[l] = {**lin[l], **{'fluo'+str(f+1):[]}}
    # Generate new label stack:
    label_stack_resized = np.empty([timepoints,chamberbox['ybr']-chamberbox['ytl'],chamberbox['xbr']-chamberbox['xtl']],dtype=np.uint16)
    for i in range(timepoints): # Go over frames
        # Resize from original stack:
        label_stack_resized[i] = cv2.resize(label_stack[i],label_stack_resized.shape[2:0:-1],interpolation=cv2.INTER_NEAREST) # The stack needs to be resized to the actual chamber box size
        # Get cropped out fluorescence images:
        if fluoframes is not None:
            chamberfluo = np.empty((fluoframes.shape[1],)+label_stack_resized.shape[1:])
            for f in range(fluoframes.shape[1]):
                chamberfluo[f] = utils.cropbox(fluoframes[i,f],chamberbox)
        # Get contours of all cells in frame:
        cells, contours = utils.getcellsinframe(label_stack_resized[i],get_contours=True)
        for c, cellnb in enumerate(cells): # Run through cells in frame
            # Length, width, area:
            rotrect = cv2.minAreaRect(contours[c])
            lin[cellnb]['length'].append(max(rotrect[1]))
            lin[cellnb]['width'].append(min(rotrect[1]))
            lin[cellnb]['area'].append(cv2.contourArea(contours[c]))
            # Pixels list and Fluorescence:
            cellpixels = np.where(label_stack_resized[i]==cellnb+1)
            lin[cellnb]['pixels'].append(np.ravel_multi_index(cellpixels,label_stack_resized.shape[1:]).astype(np.float32)) # Using floats for compatibility with previous version of the pipeline
            if fluoframes is not None:
                for f in range(fluoframes.shape[1]): #Go over all channels:
                    lin[cellnb]['fluo'+str(f+1)].append(np.mean(chamberfluo[f,cellpixels[0],cellpixels[1]]))
            
        # Check if some cells got shrunk out of existence after resize:
        cells_orig = utils.getcellsinframe(label_stack[i],get_contours=False)
        missing_cells = set(cells_orig)-set(cells)
        for cellnb in missing_cells:
            lin[cellnb]['length'].append(np.nan)
            lin[cellnb]['width'].append(np.nan)
            lin[cellnb]['area'].append(np.nan)
            lin[cellnb]['pixels'].append([])
            if fluoframes is not None:
                for f in range(fluoframes.shape[1]): #Go over all channels:
                    lin[cellnb]['fluo'+str(f+1)].append(np.nan)
            
            
                        
    return lin, label_stack_resized

### Main
if __name__=='__main__':
    
    # Default parameters:
    resfolder = None
    bioformats = False
    rotation_correction = cfg.rotation_correction
    prototype = None
    filenamesindexing = 1
    fileorder = 'pct'
    verbose = 1
    
    if len(sys.argv) >= 3: # If command line arguments were passed
        xpfolder = sys.argv[1]
        resfolder = sys.argv[2]
        i = 3
        while i<len(sys.argv):
            if sys.argv[i]=='--bio-formats':
                bioformats = bool(int(sys.argv[i+1]))
                i+=2
            if sys.argv[i]=='--order':
                fileorder = sys.argv[i+1]
                i+=2
            if sys.argv[i]=='--index':
                filenamesindexing = int(sys.argv[i+1])
                i+=2
            if sys.argv[i]=='--proto':
                prototype = sys.argv[i+1]
                i+=2
            if sys.argv[i]=='--rot':
                rotation_correction = float(sys.argv[i+1])
                i+=2
            if sys.argv[i]=='--verbose':
                verbose = int(sys.argv[i+1])
                i+=2
    
    else: # Interactive session:
        
        # Get xp settings:
        print('Experiment type?\n1 - Bio-Formats compatible (.nd2, .oib, .czi, .ome.tif...)\n2 - bioformats2sequence (folder)\n3 - micromanager (folder)\n4 - high-throughput (folder)\n0 - other (folder)\nEnter a number: ',end='')
        answer = int(input())
        print()
        
        # If bioformats file(s):
        if answer is None or answer == 1:
            print('Please select experiment file(s)...')
            xpfile = utils.getxppathdialog(ask_folder=False)
            bioformats = True
        # If folder: 
        else:
            print('Please select experiment folder...')
            xpfile = utils.getxppathdialog(ask_folder=True)
            bioformats = False
            if answer is None or answer == 2:
                prototype = None
                fileorder = 'pct'
                filenamesindexing=1
            elif answer == 3:
                prototype = 'Pos%01d/img_channel%03d_position%03d_time%09d_z000.tif'
                fileorder = 'pcpt'
                filenamesindexing=0
            elif answer == 4:
                prototype = 'chan%02d_img/Position%06d_Frame%06d.tif'
                fileorder = 'cpt'
                filenamesindexing=1
            elif answer == 0:
                print('Enter files prototype: ', end='')
                prototype = input()
                print()
                print('Enter files order: ', end='')
                fileorder = input()
                print()
                print('Enter files indexing: ', end='')
                filenamesindexing = int(input())
                print()
            else:
                raise ValueError('Invalid experiment type')
            print()
    
    # Run:
    process_xp(xpfile,
               use_bioformats=bioformats,
               prototype=prototype,
               fileorder=fileorder,
               filenamesindexing=filenamesindexing,
               resfolder = resfolder,
               rotation_correction = rotation_correction,
               write_movie = cfg.write_mp4_movies,
               verbose = verbose)